/**************************************************************************
*  Copyright (c) 2013 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  19.05.2013  mifi  First Version.
*  19.08.2013  mifi  Some rework and cleanup.
*  18.04.2014  mifi  Some rework, HW init of ETH will be done by
*                    ethernetif_init.
*  07.02.2015  mifi  Some rework and cleanup. 
**************************************************************************/
#define __IPSTACK_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include <string.h>
#include "project.h"
#include "tal.h"
#include "ipstack.h"
#include "terminal.h"

#include "lwip\api.h"
#include "lwip\tcpip.h"
#include "lwip\inet.h"
#include "lwip\netif.h"

/*=======================================================================*/
/*  Extern                                                               */
/*=======================================================================*/

extern err_t ethernetif_init (struct netif *netif);

/*=======================================================================*/
/*  Global                                                               */
/*=======================================================================*/

int errno = 0; /* Needed by lwIP */

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

static struct netif   NetIf;
static struct netif *pNetIf = NULL;

static uint8_t    MACDefaultAddr[6] = IP_DEFAULT_MAC_ADDR;
static uint8_t    MACAddr[6] = {0,0,0,0,0,0};
static uint32_t   IPAddr  = 0;
static uint32_t   IPAddrLast = 0;
static uint32_t   NETMask = 0;
static uint32_t   GWAddr  = 0;

static int        NetIfIsUp = 0;
static int        LinkIsUp  = 0;

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  atoh                                                                 */
/*                                                                       */
/*  Convert a ASCII internet address into host order.                    */
/*                                                                       */
/*  In    : pAsciiAddr                                                   */
/*  Out   : none                                                         */
/*  Return: host order address                                           */
/*************************************************************************/
uint32_t atoh (char *pAsciiAddr)
{
   uint32_t addr = inet_addr(pAsciiAddr);
   
   return( ntohl(addr) );
} /* atoh */

/*************************************************************************/
/*  htoa                                                                 */
/*                                                                       */
/*  Convert a host IP address into decimal dotted ASCII representation.  */
/*                                                                       */
/*  In    : Addr, pBuf, BufLen                                           */
/*  Out   : none                                                         */
/*  Return: ASCII address                                                */
/*************************************************************************/
char *htoa (uint32_t HostAddr, char *pBuf, int BufLen)
{
   ip_addr_t ipaddr;

   ipaddr.addr = htonl(HostAddr);

   return( ipaddr_ntoa_r(&ipaddr, pBuf, BufLen) );
} /* htoa */

/*************************************************************************/
/*  NetIfStatusCallback                                                  */
/*                                                                       */
/*  Callback is called whenever an interface changes its up/down status  */ 
/*  (i.e., due to DHCP IP acquistion)                                    */
/*                                                                       */
/*  In    : netif                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void NetIfStatusCallback (struct netif *netif)
{
   if(netif_is_up(netif))
   {
      TAL_PRINTF("NetIf is up\r\n");
      NetIfIsUp = 1;

      /*
       * Check if ip address has changed, e.g. by DHCP bound
       */      
      if (netif->ip_addr.addr != IPAddrLast)
      {
         /* DHCP bound */
         IPAddrLast = netif->ip_addr.addr;

         _IPDhcpCallbackBound();
      }
   }
   else
   {
      TAL_PRINTF("NetIf is down\r\n");
      NetIfIsUp = 0;
   }
} /* NetIfStatusCallback */

/*************************************************************************/
/*  NetIfLinkCallback                                                    */
/*                                                                       */
/*  Callback is called when link is brought up/down                      */
/*                                                                       */
/*  In    : netif                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void NetIfLinkCallback (struct netif *netif)
{
   if (netif_is_link_up(netif))
   {
      TAL_PRINTF("Link is up\r\n");
      LinkIsUp = 1;
   }
   else
   {
      TAL_PRINTF("Link is down\r\n");
      LinkIsUp = 0;
   }
} /* netif_link_callback */

/*************************************************************************/
/*  IPIfaceMACSet                                                        */
/*                                                                       */
/*  Set the MAC address of the network interface.                        */
/*                                                                       */
/*  In    : mac                                                          */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void IPIfaceMACSet (uint8_t *mac)
{
   memcpy(MACAddr, mac, 6);
} /* IPIfaceMACSet */

/*************************************************************************/
/*  IPIfaceAddrSet                                                       */
/*                                                                       */
/*  Set the IP address of the network interface in host order.           */
/*                                                                       */
/*  In    : addr                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void IPIfaceAddrSet (uint32_t addr)
{
   IPAddr = htonl(addr);
} /* IPIfaceAddrSet */

/*************************************************************************/
/*  IPIfaceMaskSet                                                       */
/*                                                                       */
/*  Set the network mask of the network interface in host order.         */
/*                                                                       */
/*  In    : addr                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void IPIfaceMaskSet (uint32_t addr)
{
   NETMask = htonl(addr);
} /* IPIfaceMaskSet */

/*************************************************************************/
/*  IPIfaceGWSet                                                         */
/*                                                                       */
/*  Set the gateway address of the network interface in host order.      */
/*                                                                       */
/*  In    : addr                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void IPIfaceGWSet (uint32_t addr)
{
   GWAddr = htonl(addr);
} /* IPIfaceGWSet */

/*************************************************************************/
/*  IPIfaceMACGet                                                        */
/*                                                                       */
/*  Get the MAC address of the network interface.                        */
/*                                                                       */
/*  In    : mac, size                                                    */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void IPIfaceMACGet (uint8_t *mac, uint8_t size)
{
   if (size == sizeof(MACAddr))
   {
      memcpy(mac, pNetIf->hwaddr, 6);
   }
} /* IPIfaceMACGet */

/*************************************************************************/
/*  IPIfaceAddrGet                                                       */
/*                                                                       */
/*  Get the IP address of the network interface in host order.           */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Address in host order                                        */
/*************************************************************************/
uint32_t IPIfaceAddrGet (void)
{
   uint32_t addr = 0;

   if (pNetIf != NULL)
   {
      addr = pNetIf->ip_addr.addr;
   }

   return(ntohl(addr));
} /* IPIfaceAddrGet */

/*************************************************************************/
/*  IPIfaceMaskGet                                                       */
/*                                                                       */
/*  Get the network mask of the network interface in host order.         */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Address in host order                                        */
/*************************************************************************/
uint32_t IPIfaceMaskGet (void)
{
   uint32_t addr = 0;

   if (pNetIf != NULL)
   {
      addr = pNetIf->netmask.addr;
   }

   return(ntohl(addr));
} /* IPIfaceMaskGet */

/*************************************************************************/
/*  IPIfaceGWGet                                                         */
/*                                                                       */
/*  Get the gateway address of the network interface in host order.      */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Address in host order                                        */
/*************************************************************************/
uint32_t IPIfaceGWGet (void)
{
   uint32_t addr = 0;

   if (pNetIf != NULL)
   {
      addr = pNetIf->gw.addr;
   }

   return(ntohl(addr));
} /* IPIfaceGWGet */

/*************************************************************************/
/*  IPIfaceStart                                                         */
/*                                                                       */
/*  Start the TCP/IP stack.                                              */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: 0 on success / -1 otherwise                                  */
/*************************************************************************/
int IPIfaceStart (void)
{
   int       Error = -1;
   ip_addr_t ipaddr;
   ip_addr_t netmask;
   ip_addr_t gw;  
   
   /* Check if IP address or network mask is 0 */ 
   if ((0 == IPAddr) || (0 == NETMask))
   {
      /* Error, use default values */
      IPAddr  = inet_addr(IP_DEFAULT_ADDRESS);
      NETMask = inet_addr(IP_DEFAULT_NETMASK);
      GWAddr  = inet_addr(IP_DEFAULT_GATEWAY);
   }

   /* Check if MAC address is 0 */
   if( (0 == MACAddr[0]) && (0 == MACAddr[1]) && (0 == MACAddr[2]) &&
       (0 == MACAddr[3]) && (0 == MACAddr[4]) && (0 == MACAddr[5]) )
   {
      /* Error, use default values */
      memcpy(MACAddr, MACDefaultAddr, sizeof(MACAddr));   
   }
   
   /* IP address setting */
   ipaddr.addr  = IPAddr;
   netmask.addr = NETMask;
   gw.addr      = GWAddr;
   
   IPAddrLast = IPAddr; 

   /* Initialize lwIP */
   tcpip_init(NULL, NULL);

   /* Add the network interface to the list of lwIP netifs. */
   pNetIf = netif_add(&NetIf, &ipaddr, &netmask, &gw, MACAddr, ethernetif_init,tcpip_input);
   if (pNetIf != NULL)
   {
      /* Set the network interface as the default network interface. */
      netif_set_default(pNetIf);
            
      /*
       * Set the callback function whenever an interface
       * changes its up/down status (i.e., due to DHCP IP acquistion)
       */
      netif_set_status_callback(&NetIf, NetIfStatusCallback);

      /* 
       * Set the callback function from an interface 
       * whenever the link changes (i.e., link down)
       */      
      netif_set_link_callback(&NetIf, NetIfLinkCallback);      

      /* Bring the interface up. */
      netif_set_up(pNetIf); 
      
      Error = 0;
   }      
   
   return(Error);
} /* IPIfaceStart */    

/*************************************************************************/
/*  IPIfaceStop                                                          */
/*                                                                       */
/*  Stop the TCP/IP stack.                                               */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: 0 on success / -1 otherwise                                  */
/*************************************************************************/
int IPIfaceStop (void)
{
   int Error = -1;
   
   if (pNetIf != NULL)
   {
      netif_set_down(pNetIf);
   }
   
   return(Error);
} /* IPIfaceStop */      

/*************************************************************************/
/*  IPIsNetIfUp                                                          */
/*                                                                       */
/*  Check if the interface and link is Up.                               */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: 0 / 1                                                        */
/*************************************************************************/
int IPIsNetIfUp (void)
{
   int NetIfUp = 0;
   
   if ((netif_is_up(&NetIf)) && (netif_is_link_up(&NetIf)))
   {
      NetIfUp = 1;
   }
   
   return(NetIfUp);
} /* IPIsNetIfUp */

/*************************************************************************/
/*  IPIfaceOutputConfig                                                  */
/*                                                                       */
/*  Display IP configuration.                                            */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void IPIfaceOutputConfig (void)
{
   char Buffer[24];
   
   term_printf("*** Network configuration ***\r\n");
   term_printf("\r\n");
   
   
   term_printf("Interface      : %s\r\n", (NetIfIsUp) ? "up" : "down");
   term_printf("Link           : %s\r\n", (LinkIsUp) ? "up" : "down");
   
   IPIfaceMACGet((uint8_t*)Buffer, sizeof(MACAddr));   
   term_printf("MAC Address    : %02X:%02X:%02X:%02X:%02X:%02X\r\n",
               Buffer[0], Buffer[1], Buffer[2], Buffer[3], Buffer[4], Buffer[5]);
               
   term_printf("DHCP Enabled   : %s\r\n", (IPDhcpIsStarted() == 0) ? "no" : "yes");
   term_printf("mDNS Enabled   : %s\n",   (IPmDnsIsStarted() == 0) ? "no" : "yes");
   term_printf("IP Address     : %s\r\n", htoa(IPIfaceAddrGet(), Buffer, sizeof(Buffer)));
   term_printf("Subnet Mask    : %s\r\n", htoa(IPIfaceMaskGet(), Buffer, sizeof(Buffer)));
   term_printf("Gateway        : %s\r\n", htoa(IPIfaceGWGet(), Buffer, sizeof(Buffer)));
   term_printf("DHCP Server    : %s\r\n", htoa(IPDhcpServerGet(), Buffer, sizeof(Buffer)));
   term_printf("\r\n");

} /* IPIfaceOutputConfig */

/*** EOF ***/
